import unittest

from artap.problem import Problem
from artap.benchmark_functions import BinhAndKorn, AckleyN2
from artap.algorithm_genetic import EpsMOEA
from artap.results import Results, GraphicalResults

# import optproblems as optp
# import optproblems.cec2005 as cec2005
# import optproblems.dtlz as dtlz

# from math import pi


# class BinhAndKornProblem(Problem):
#
#     """ Describe simple one objective optimization problem. """
#     def set(self):
#         self.name = "EPSMOEA"
#         self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [0, 5]},
#                            {'name': 'x_2', 'initial_value': 1.5, 'bounds': [0, 3]}]
#         self.costs = [{'name': 'F_1'}, {'name': 'F_2'}]
#
#     def evaluate(self, x):
#         function = BinhAndKorn()
#         return function.eval(x)
#
#     def evaluate_constraints(self, x):
#         return BinhAndKorn.constraints(x)
#
#
# class TestEPSMOEAOptimization(unittest.TestCase):
#     """ Tests simple one objective optimization problem."""
#
#     def test_local_problem_epsmoea(self):
#         problem = BinhAndKornProblem()
#         algorithm = EpsMOEA(problem)
#         algorithm.options['max_population_number'] = 100
#         algorithm.options['max_population_size'] = 100
#         algorithm.options['calculate_gradients'] = True
#         algorithm.options['verbose_level'] = 1
#         algorithm.options['epsilons'] = 0.05
#
#         algorithm.run()
#
#         b = Results(problem)
#         solution = b.pareto_values()
#         wrong = 0
#         for sol in solution:
#             if abs(BinhAndKorn.approx(sol[0]) - sol[1]) > 0.1 * BinhAndKorn.approx(sol[0]) \
#                     and 20 < sol[0] < 70:
#                 wrong += 1
#
#         self.assertLessEqual(wrong, 5)


class AckleyN2Test(Problem):
    """Test the convergence in a one objective example with a simple 2 variable Ackley N2 formula"""

    def set(self):
        self.name = "TestAckleyN2"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-32, 32]},
                      {'name': 'x_2', 'initial_value': 2.5, 'bounds': [-32, 32]}]
        self.costs = [{'name': 'F_1'}]

    def evaluate(self, individual):
        function = AckleyN2()
        return [function.eval(individual.vector)]


class TestAckleyN222(unittest.TestCase):
    """ Tests that the eps-moea algorithm can find the global optimum of a function."""

    def test_local_problem(self):
        problem = AckleyN2Test()
        algorithm = EpsMOEA(problem)
        algorithm.options['max_population_number'] = 100
        algorithm.options['max_population_size'] = 100
        algorithm.options['epsilons'] = 0.01
        algorithm.run()

        b = Results(problem)
        optimum = b.find_minimum('F_1')  # Takes last cost function
        self.assertAlmostEqual(optimum.costs[0], -200, 0)
#####
#####
# class CEC2005_TEST_Problems(Problem):
#
#     # these problems working for 2, 10, 30 variables in every case
#
#     def __init__(self, name, test_function, ub, lb):
#
#         #ub = 100
#         #lb = -100
#
#         parameters = {'x_1', 'initial_value': 2.5, 'bounds': [lb, ub]},
#                       'x_2', 'initial_value': 2.5, 'bounds': [lb, ub]}}
#
#         costs = ['F_1']
#         self.function = test_function(2)
#
#         super().__init__(name, parameters, costs)
#
#
#     def evaluate(self, x):
#
#         #function = optp.cec2005.F1(2)
#         problem = optp.Problem(self.function, num_objectives = 1)
#         solutions = [optp.Individual(x)]
#
#         problem.batch_evaluate(solutions)
#
#         return [solutions[0].objective_values]

# class TestCEC2005(unittest.TestCase):
#     """ Tests that the NSGA II algorithm can find the global optimum of a function."""
#
#     def run_test_problem(self, method, ub, lb, gen_nr):
#         problem = CEC2005_TEST_Problems("CEC2005",method, ub, lb)
#         algorithm = NSGAII(problem)
#         algorithm.options['max_population_number'] = gen_nr
#         algorithm.options['max_population_size'] = 100
#         algorithm.run()
#
#         b = Results(problem)
#         optimum = b.find_minimum('F_1')  # Takes last cost function
#         self.assertAlmostEqual(optimum.costs[0], method.bias, 0)
#
#     # unimodal problems
#
#     def test_shifted_sphere(self):
#         self.run_test_problem(optp.cec2005.F1, 100, -100, 100)
#
#     def test_shifted_double_sum(self):
#         self.run_test_problem(optp.cec2005.F2, 100, -100, 100)
#
#     # def test_shifted_rotated_elliptic(self): # wrong with platypus also
#     #     self.run_test_problem(optp.cec2005.F3, 100, -100, 10000)
#     #
#     #def test_shifted_schwefel(self): # wrong with platypus also
#     #    self.run_test_problem(optp.cec2005.F4, 100, -100, 10000)
#
#     def test_schwefel(self): # !
#         self.run_test_problem(optp.cec2005.F5, 100, -100, 100)
#
#     def test_shifted_rosenbrock(self):
#         self.run_test_problem(optp.cec2005.F6, 100, -100, 1000)
#
#     # multimodal problems
#     #def test_shifted_rot_griewank(self):
#     #    self.run_test_problem(optp.cec2005.F7, 100, -100, 10000)
#
#     #def test_shifted_rot_ackley(self):
#     #    self.run_test_problem(optp.cec2005.F8, 32, -32, 100)
#
#     def test_shifted_rastrigin(self):
#         self.run_test_problem(optp.cec2005.F9, 5, -5, 100)
#
#     def test_shifted_rot_rastrigin(self):
#         self.run_test_problem(optp.cec2005.F10, 5, -5, 1000)
#
#     def test_shifted_rot_weierstrass(self):
#         self.run_test_problem(optp.cec2005.F11, 0.5, -0.5, 100)
#
#     def test_fletcherpowell(self): #schwefel
#         self.run_test_problem(optp.cec2005.F12, pi, -pi, 100)
#
#     def test_f8f2(self):
#         self.run_test_problem(optp.cec2005.F13, 1, -3, 100)
#
#     def test_expanded_f6(self):
#         self.run_test_problem(optp.cec2005.F14, 100, -100, 100)
#
#     #####
#     def test_hybrid_composition(self):
#         self.run_test_problem(optp.cec2005.F15, 5, -5, 100)
#
#     def test_rotated_f15(self):
#         self.run_test_problem(optp.cec2005.F16, 5, -5, 100)
#
#     def test_f16_with_noise(self):
#         self.run_test_problem(optp.cec2005.F17, 5, -5, 250)
#
#     # def test_rot_hybrid_function(self):
#     #     self.run_test_problem(optp.cec2005.F18, 5, -5, 1000)
#     #
#     # def test_rot_hybrid_function_2(self):
#     #     self.run_test_problem(optp.cec2005.F19, 5, -5, 1000)
#     #
#     # def test_rot_hybrid_function_3(self):
#     #     self.run_test_problem(optp.cec2005.F20, 5, -5, 1000)
#     #
#     # def test_rot_hybrid_function_4(self):
#     #     self.run_test_problem(optp.cec2005.F21, 5, -5, 1000)
#     #
#     # def test_rot_hybrid_function_5(self):
#     #     self.run_test_problem(optp.cec2005.F22, 5, -5, 1000)
#     #
#     # def test_rot_hybrid_function_6(self):
#     #     self.run_test_problem(optp.cec2005.F23, 5, -5, 1000)
#     #
#     # def test_rot_hybrid_function_7(self):
#     #     self.run_test_problem(optp.cec2005.F24, 5, -5, 1000)
#     #
#     # def test_rot_hybrid_function_wo_bounds(self):
#     #     self.run_test_problem(optp.cec2005.F25, 100, -100, 100)


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase, main

from artap.executor import CondorComsolJobExecutor, CondorMatlabJobExecutor, CondorPythonJobExecutor
from artap.problem import Problem, ProblemType
from artap.population import Population
from artap.algorithm import DummyAlgorithm


class CondorMatlabProblem(Problem):
    """ Describe simple one objective optimization problem. """

    def set(self):
        self.name = "CondorMatlabProblem"
        self.parameters = [{'name': 'a', 'initial_value': 10, 'bounds': [0, 20]},
                      {'name': 'b', 'initial_value': 10, 'bounds': [5, 15]}]
        self.costs = [{'name': 'F_1'}]
        self.type = ProblemType.matlab
        self.executor = CondorMatlabJobExecutor(self,
                                                script="./data/run_input.m",
                                                parameter_file="input.txt",
                                                files_from_condor=["output.txt"])

    def evaluate(self, individual):
        return self.executor.eval(individual)    # ToDo: could be passed individual?

    def parse_results(self, output_files, individual):
        with open(output_files[0]) as file:
            content = file.readlines()
        return [float(content[0])]


class CondorComsolProblem(Problem):
    """ Describe simple one objective optimization problem. """

    def set(self):
        self.name = "CondorComsolProblem"
        self.parameters = [{'name': 'a', 'initial_value': 10, 'bounds': [0, 20]},
                           {'name': 'b', 'initial_value': 10, 'bounds': [5, 15]}]

        self.costs = [{'name': 'F_1'}]
        self.type = ProblemType.comsol
        self.executor = CondorComsolJobExecutor(self, model_file="./data/elstat.mph",
                                                files_from_condor=["out.txt", "elstat.mph"], )

    def evaluate(self, individual):
        return self.executor.eval(individual)

    def parse_results(self, output_files, individual):
        with open(output_files[0]) as file:
            content = file.readlines()
        line_with_results = content[5]  # 5th line contains results
        result = float(line_with_results)
        return [result]


class PythonExecProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.parameters = [{'name': 'a', 'initial_value': 10, 'bounds': [0, 20]},
                           {'name': 'b', 'initial_value': 10, 'bounds': [5, 15]}]
        self.costs = [{'name': 'F_1'}]
        self.type = ProblemType.python
        self.executor = CondorPythonJobExecutor(self,
                                                script="./data/run_exec.py",
                                                parameter_file=None,
                                                output_files=["output.txt"])

    def evaluate(self, individual):
        return self.executor.eval(individual)

    def parse_results(self, output_files, individual):
        with open(output_files[0]) as file:
            content = file.readlines()
        return [float(content[0])]


class PythonInputProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.parameters = [{'name': 'a', 'initial_value': 10, 'bounds': [0, 20]},
                           {'name': 'b', 'initial_value': 10, 'bounds': [5, 15]}]
        self.costs = [{'name': 'F_1'}]
        self.type = ProblemType.python
        self.executor = CondorPythonJobExecutor(self,
                                                script="./data/run_input.py",
                                                parameter_file="input.txt",
                                                output_files=["output.txt"])

    def evaluate(self, individual):
        return self.executor.eval(individual)

    def parse_results(self, output_files, individual):
        with open(output_files[0]) as file:
            content = file.readlines()
        return [float(content[0])]


class TestCondor(TestCase):
    """ Tests simple optimization problem where calculation of
        goal function is submitted as a job on HtCondor.
    """
    def test_condor_matlab_input(self):
        """ Tests one calculation of goal function."""
        problem = CondorMatlabProblem()

        table = [[1, 2]]
        population = Population()
        population.gen_population_from_table(table)
        evaluator = DummyAlgorithm(problem)
        evaluator.evaluate(population.individuals)

        self.assertAlmostEqual(5, population.individuals[0].costs[0])

    def test_condor_comsol_exec(self):
        """ Tests one calculation of goal function."""
        problem = CondorComsolProblem()
        problem.options['save_data_files'] = True
        table = [[10, 10], [11, 11]]
        population = Population()
        population.gen_population_from_table(table)
        evaluator = DummyAlgorithm(problem)
        evaluator.evaluate(population.individuals)

        self.assertAlmostEqual(112.94090668383139, population.individuals[0].costs[0])
        self.assertAlmostEqual(124.23499735221547, population.individuals[1].costs[0])

    def test_condor_python_exec(self):
        problem = PythonExecProblem()

        table = [[1, 2]]
        population = Population()
        population.gen_population_from_table(table)
        evaluator = DummyAlgorithm(problem)
        evaluator.evaluate(population.individuals)

        self.assertAlmostEqual(5, population.individuals[0].costs[0])

    def test_condor_python_input(self):
        problem = PythonInputProblem()

        table = [[1, 2]]
        population = Population()
        population.gen_population_from_table(table)
        evaluator = DummyAlgorithm(problem)
        evaluator.evaluate(population.individuals)

        self.assertAlmostEqual(5, population.individuals[0].costs[0])


if __name__ == '__main__':
    main()



__version__ = "0.1.9"
__author__ = u"Artap Team"
__author_email__ = "artap.team@gmail.com"
__copyright__ = u"Copyright (c) 2018-2019, {} <{}>".format(__author__, __author_email__)
__website__ = "https://edison.fel.zcu.cz:444/agros/RDOLab/Artap"
__license__ = "License :: OSI Approved :: MIT License"
__status__ = "Development Status :: 3 - Alpha"



import unittest
import sys

from artap.problem import Problem
from artap.algorithm_nlopt import NLopt
from artap.algorithm_nlopt import GN_DIRECT_L
# from artap.algorithm_nlopt import GN_DIRECT_L_RAND
from artap.algorithm_nlopt import GN_MLSL
from artap.algorithm_nlopt import GN_CRS2_LM
from artap.algorithm_nlopt import GN_ISRES
# from artap.algorithm_nlopt import GN_ESCH
from artap.algorithm_nlopt import LN_BOBYQA
from artap.algorithm_nlopt import LN_COBYLA
from artap.algorithm_nlopt import LN_NELDERMEAD
from artap.algorithm_nlopt import LN_SBPLX
from artap.algorithm_nlopt import LN_PRAXIS
from artap.algorithm_nlopt import LN_AUGLAG_EQ

from artap.results import Results
from artap.benchmark_functions import Booth
from agrossuite import agros as a2d


class MyProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.name = "AgrosProblem"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-10, 10]},
                      {'name': 'x_2', 'initial_value': 1.5, 'bounds': [-10, 10]}]
        self.costs = [{'name': 'F'}]

    def evaluate(self, individual):
        return [Booth.eval(individual.vector)]


class TestNLoptOptimization(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def run_test(self, method, n_iterations=100):
        try:
            problem = MyProblem()
            algorithm = NLopt(problem)
            algorithm.options['verbose_level'] = 0
            algorithm.options['algorithm'] = method
            algorithm.options['xtol_abs'] = 1e-6
            algorithm.options['xtol_rel'] = 1e-3
            algorithm.options['ftol_rel'] = 1e-3
            algorithm.options['ftol_abs'] = 1e-6
            algorithm.options['n_iterations'] = n_iterations
            algorithm.run()

            results = Results(problem)
            optimum = results.find_minimum('F')
            self.assertAlmostEqual(optimum.costs[0], 0, places=1)
        except AssertionError:
            # try again with more iterations - NLopt - 3rd party library
            self.run_test(method, int(1.5 * n_iterations))

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_local_problem_nlopt_GN_DIRECT_L(self):
        self.run_test(GN_DIRECT_L, 400)

    # @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    # def test_local_problem_nlopt_GN_DIRECT_L_RAND(self):
    #    self.run_test(GN_DIRECT_L_RAND, 100)

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_local_problem_nlopt_GN_MLSL(self):
        self.run_test(GN_MLSL, 500)

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_local_problem_nlopt_GN_CRS2_LM(self):
        self.run_test(GN_CRS2_LM, 1000)

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_local_problem_nlopt_GN_ISRES(self):
        self.run_test(GN_ISRES, 1000)

    # @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    # def test_local_problem_nlopt_GN_ESCH(self):
    #     self.run_test(GN_ESCH, 5000)

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_local_problem_nlopt_LN_BOBYQA(self):
        self.run_test(LN_BOBYQA)

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_local_problem_nlopt_LN_COBYLA(self):
        self.run_test(LN_COBYLA)

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_local_problem_nlopt_LN_NELDERMEAD(self):
        self.run_test(LN_NELDERMEAD)

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_local_problem_nlopt_LN_SBPLX(self):
        self.run_test(LN_SBPLX, 200)

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_local_problem_nlopt_LN_AUGLAG_EQ(self):
        self.run_test(LN_AUGLAG_EQ)

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_local_problem_nlopt_LN_PRAXIS(self):
        self.run_test(LN_PRAXIS)


class SimpleCapacitorProblem(Problem):
    """
    There was a bug between the job queue and the nlopt --- this is a testproblem to test the interface between the
    agros and artap

    """
    def set(self):
        self.parameters = [{'name': 'r1', 'initial_value': 0.02, 'bounds': [0.01, 0.03]},
                      {'name': 'r2', 'initial_value': 0.04, 'bounds': [0.035, 0.1]}]
        self.costs = [{'name': 'F'}]

    def evaluate(self, individual):
        x = individual.vector
        C_req = 80.  # pF
        # problem
        problem = a2d.problem(clear=True)
        problem.coordinate_type = "planar"
        problem.mesh_type = "triangle"

        # project parameters
        problem.parameters["r1"] = x[0]
        problem.parameters["r2"] = x[1]
        problem.parameters["eps"] = 2.5

        # fields
        # electrostatic
        electrostatic = problem.field("electrostatic")
        electrostatic.analysis_type = "steadystate"
        electrostatic.matrix_solver = "umfpack"
        electrostatic.number_of_refinements = 1
        electrostatic.polynomial_order = 2
        electrostatic.adaptivity_type = "disabled"
        electrostatic.solver = "linear"

        # boundaries
        electrostatic.add_boundary("GND", "electrostatic_potential", {"electrostatic_potential": 0})
        electrostatic.add_boundary("V", "electrostatic_potential", {"electrostatic_potential": 1})
        electrostatic.add_boundary("neumann", "electrostatic_surface_charge_density",
                                   {"electrostatic_surface_charge_density": 0})

        # materials
        electrostatic.add_material("dielectric",
                                   {"electrostatic_permittivity": "eps", "electrostatic_charge_density": 0})

        # geometry
        geometry = problem.geometry()
        geometry.add_edge("r1", 0, "r2", 0, boundaries={"electrostatic": "neumann"})
        geometry.add_edge("r2", 0, 0, "r2", angle=90, boundaries={"electrostatic": "GND"})
        geometry.add_edge(0, "r2", 0, "r1", boundaries={"electrostatic": "neumann"})
        geometry.add_edge("r1", 0, 0, "r1", angle=90, boundaries={"electrostatic": "V"})

        geometry.add_label("(r1 + r2) / 2.", "(r2 - r1) / 2", materials={"electrostatic": "dielectric"})

        computation = problem.computation()
        computation.solve()

        solution = computation.solution("electrostatic")
        result = solution.volume_integrals()["We"]

        C = 4 * 2 * result *1e12 # pF

        return [abs(C - C_req)]


class TestSimpleCapacitor(unittest.TestCase):

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_run(self):
        problem = SimpleCapacitorProblem()
        algorithm = NLopt(problem)
        algorithm.options['n_iterations'] = 300
        algorithm.options['algorithm'] = LN_BOBYQA
        algorithm.options['verbose_level'] = 0
        algorithm.run()

        results = Results(problem)
        optimum = results.find_minimum('F')

        self.assertAlmostEqual(optimum.costs[0], 0, places=1)
        # tests that the optimal value is within the predefined boundaries
        # --- excludes possible errors between -> job -- nlopt and memory handling

        # r1 between the limits?
        self.assertTrue(optimum.vector[0] >= problem.parameters[0]['bounds'][0])
        self.assertTrue(optimum.vector[0] <= problem.parameters[0]['bounds'][1])

        # r2 between the limits?
        self.assertTrue(optimum.vector[1] >= problem.parameters[1]['bounds'][0])
        self.assertTrue(optimum.vector[1] <= problem.parameters[1]['bounds'][1])


if __name__ == '__main__':
    unittest.main()



from .problem import Problem
from .algorithm import Algorithm
from .population import Population
from .individual import Individual
import collections
from .job import JobSimple


class GradientDescent(Algorithm):
    """ Gradient descent algorithms """

    def __init__(self, problem: Problem, name="GradientDescent"):
        super().__init__(problem, name)
        self.problem = problem

        self.options.declare(name='n_iterations', default=10,
                             desc='Maximum number of iterations')

        self.options.declare(name='x0', default=0,
                             desc='initial point')

        self.options.declare(name='h', default=0.1,
                             desc='step')

    def evaluate_gradient_richardson(self, population, individual):
        x0 = individual.vector
        gradient = [0] * len(x0)

        h = 1e-6
        job = JobSimple(self.problem, population)
        y = job.evaluate_scalar(x0)
        for i in range(len(x0)):
            x = x0.copy()
            x[i] += h
            y_h = job.evaluate_scalar(x)
            d_0_h = gradient[i] = (y_h - y) / h
            x[i] += h
            y_2h = job.evaluate_scalar(x)
            d_0_2h = (y_2h - y) / 2 / h
            gradient[i] = (4 * d_0_h - d_0_2h) / 3

        return gradient

    def evaluate_gradient(self, individual):
        x0 = individual.vector
        gradient = [0] * len(x0)
        x0 = individual.parameters
        y = self.evaluate(x0)
        h = 1e-6
        for i in range(len(x0)):
            x = x0.copy()
            x[i] += h
            y_h = self.evaluate(x)
            if isinstance(y_h, collections.Iterable):
                m = len(y_h)
                gradient[i] = []
                for j in range(m):
                    gradient[i].append((y_h[j] - y[j]) / h)
            else:
                gradient[i] = (y_h - y) / h

        return gradient

    def run(self):
        population = Population()

        # TODO: add adaptive step size
        n = self.options["n_iterations"]
        x = [] * n
        dx = [] * n
        x.append(self.options["x0"])
        dx.append([0, 0])
        h = self.options["h"]
        for i in range(1, n):
            x.append([])
            dx.append([])
            individual = Individual(x[i-1])
            gradient = self.evaluate_gradient_richardson(population, individual)
            dx[i] = gradient

            for j in range(len(x[i-1])):
                x[i].append(x[i-1][j] - h * gradient[j])

            if i > 1:
                d = 0
                n = 0
                for j in range(len(x[i])):
                    n += (dx[i][j] - dx[i-1][j]) * (x[i][j] - x[i-1][j])
                    d += (dx[i][j] - dx[i-1][j])**2
                h = n / d
            population.individuals.append(individual)

        # write to datastore
        self.problem.data_store.write_population(population)



import unittest

from artap.problem import Problem
from artap.algorithm_scipy import ScipyOpt
from artap.benchmark_functions import AckleyN2
from artap.results import Results


class MyProblem(Problem):
    """ Describe simple one objective optimization problem. """

    def set(self):
        self.name = "TestSimpleOptimization"
        self.parameters = [{'name': 'x_1', 'initial_value': 10}]
        self.costs = [{'name': 'F_1'}]

    def evaluate(self, individual):
        x = individual.vector
        result = 0
        for i in x:
            result += i*i

        return [result]


class TestSimpleOptimization(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_local_problem(self):
        problem = MyProblem()
        algorithm = ScipyOpt(problem)
        algorithm.options['algorithm'] = 'Nelder-Mead'
        algorithm.options['tol'] = 1e-4
        algorithm.run()

        results = Results(problem)
        optimum = results.find_minimum('F_1')
        self.assertAlmostEqual(optimum.costs[0], 0)


class AckleyN2Problem(Problem):
    """Test with a simple 2 variable Ackley N2 formula"""

    def set(self):
        self.name = "AckleyN2"
        self.parameters = {'x': {'initial_value': 2.13}, 'y': {'initial_value': 2.13}}
        self.costs = [{'name': 'F_1'}]

    def evaluate(self, individual):
        function = AckleyN2()
        return [function.eval(individual.vector)]


class TestAckleyN2(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def test_local_problem(self):
        problem = AckleyN2Problem()
        algorithm = ScipyOpt(problem)
        algorithm.options['algorithm'] = 'Nelder-Mead'
        algorithm.options['tol'] = 1e-4
        algorithm.options['calculate_gradients'] = True
        algorithm.run()

        results = Results(problem)
        optimum = results.find_minimum('F_1')
        self.assertAlmostEqual(optimum.costs[0], -200, 3)


if __name__ == '__main__':
    unittest.main()



rom artap.problem import Problem
from artap.algorithm_genetic import NSGAII, EpsMOEA
from artap.results import Results, GraphicalResults

import matplotlib.pyplot as plt


class SpringDesignProblem(Problem):
    """
    Example from K.DEb Multi-objective evolutionary optimization problems, Wiley, 2001.
    pp 435.


    TODO: integer based problem, later


    """
    def __init__(self, name):
        parameters = {'x_1': {'initial_value': .5, 'bounds': [0.1, 1.]},
                      'x_2': {'initial_value': .5, 'bounds': [0.0, 5.0]}}

        costs = ['F_1', 'F_2']

        super().__init__(name, parameters, costs)

    def evaluate(self, x):
        f1 =
        f2 =
        return [f1, f2]


from multiprocessing import Process, Pipe

def f(conn):
    conn.send([42, None, 'hello'])
    conn.close()

if __name__ == '__main__':
    parent_conn, child_conn = Pipe()
    p = Process(target=f, args=(child_conn,))
    p.start()
    print(parent_conn.recv())   # prints "[42, None, 'hello']"
    print(child_conn.recv())
    p.join()


import unittest
import sys

from artap.problem import Problem
from artap.algorithm_bayesopt import BayesOptSerial, BayesOptParallel

from artap.results import Results
from artap.benchmark_functions import Booth


class MyProblem(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.name = "TestBayesOptParallel"
        self.parameters = [{'name': 'x_1', 'initial_value': 2.5, 'bounds': [-10, 10]},
                      {'name': 'x_2', 'initial_value': 1.5, 'bounds': [-10, 10]}]
        self.costs = [{'name': 'F'}]

    def evaluate(self, individual):
        return [Booth.eval(individual.vector)]


class TestBayesOptOptimization(unittest.TestCase):
    """ Tests simple one objective optimization problem."""

    def xtest_local_problem_bayesopt_parallel(self):
        problem = MyProblem()
        algorithm = BayesOptParallel(problem)
        algorithm.options['verbose_level'] = 0
        algorithm.options['n_iterations'] = 100
        algorithm.run()
        # TODO - multiprocess test

        results = Results(problem)
        optimum = results.find_minimum(name='F')
        self.assertAlmostEqual(optimum.costs[0], 0, places=2)

    @unittest.skipUnless(sys.platform.startswith("linux"), "requires linux")
    def test_local_problem_bayesopt_serial(self):
        problem = MyProblem()
        algorithm = BayesOptSerial(problem)
        algorithm.options['verbose_level'] = 0
        algorithm.options['n_iterations'] = 200
        algorithm.run()

        results = Results(problem)
        optimum = results.find_minimum(name='F')
        self.assertAlmostEqual(optimum.costs[0], 0, places=2)


if __name__ == '__main__':
    unittest.main()



from .utils import VectorAndNumbers
from .algorithm import Algorithm
from .population import Population
from .individual import Individual


class Sensitivity(Algorithm):
    def __init__(self, problem, parameters, name='Sensitivity analysis'):
        self.parameters = parameters
        super().__init__(problem, name)

        self.options.declare(name='max_population_size', default=100, lower=1,
                             desc='Maximal number of individuals in population')

    def run(self):
        self.population_size = self.options['max_population_size']

        parameters = []
        for parameter in self.problem.parameters:
            parameters.append(float(parameter['initial_value']))

        population = Population()

        for parameter_name in self.parameters:
            parameter_values = []

            index = 0
            selected_parameter = None
            for parameter in self.parameters:
                if parameter['name'] == parameter_name['name']:
                    selected_parameter = parameter
                    break
                index += 1

            individuals = []
            for i in range(self.options['max_population_size']):
                value = VectorAndNumbers.gen_number(selected_parameter['bounds'], selected_parameter['precision'],
                                              'normal')
                parameters[index] = value
                parameter_values.append(value)
                individual = Individual(parameters.copy())
                individuals.append(individual)

            self.evaluate(population.individuals)
            costs = []
            # TODO: Make also for multi-objective
            for individual in population.individuals:
                costs.append(individual.costs)

        # write to datastore
        self.problem.data_store.write_population(population)



import setuptools
from setuptools.command.build_ext import build_ext as _build_ext
import os


base_dir = os.path.abspath(os.path.dirname(__file__))
about = {}
with open(os.path.join(base_dir, "artap", "__about__.py"), "rb") as f:
    exec(f.read(), about)

with open("README.md", "r") as fh:
    long_description = fh.read()

class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

setuptools.setup(
    name="artap",
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    packages=setuptools.find_packages(),
    description="Platform for robust design optimization",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://www.agros2d.org/artap/",
    python_requires='>3.6',
    license=about["__license__"],
    data_files=[('artap', ['artap/environment.json']), ('artap/lib', ['artap/lib/bayesopt.so']),
                ('artap/lib', ['artap/lib/_nlopt.so'])],
    install_requires=['numpy',
                      'scipy',
                      'dash>=0.39.0',
                      'dash-core-components>=0.44.0',
                      'dash-html-components>=0.14.0',
                      'dash-table>=3.4.0',
                      'sklearn',
                      "paramiko",
                      'matplotlib',
                      'optproblems',
                      'smt'],
    scripts=['3rdparty/submodules.sh'],
    classifiers=[
        "Intended Audience :: Science/Research",
        "Operating System :: POSIX :: Linux",
        "Topic :: Scientific/Engineering",
        'Programming Language :: Python :: 3.6',
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb  8 13:20:29 2019

@author: david
"""

import numpy as np
import pylab as pl


class Model:
    def __init__(self, dt=0.0):
        self.t = 0
        self.type = ""
        self.dt = dt  # Sampling period
        self.systems = []

    def run(self, t):
        n = int(t / self.dt)
        for i in range(n):
            for sub_system in self.systems:
                if sub_system.type == 'discrete':
                    if abs(self.t - sub_system.k * sub_system.dt) < 1e-12:
                        sub_system.discrete_step()
                else:
                    sub_system.continuous_step(self.dt)
            self.t += self.dt


class System:
    def __init__(self, model=None, x_0=None, dt=0, n_inputs=1, n=1):

        self.n = n
        if model is not None:
            self.model = model
            self.model.systems.append(self)

        if x_0 is None:
            self.x = np.zeros(n)
        else:
            self.x = x_0

        self.t = 0
        self.k = 0
        self.dt = dt
        self.y = []                     # current output
        self.u = np.zeros(n_inputs)     # current input

        self.inputs = []
        self.connections = []
        for i in range(n_inputs):
            self.inputs.append(self.dummy_input)
            self.connections.append(0)

        self.archive_t = []
        self.archive_y = []
        self.archive_x = []
        self.archive_u = []

        self.integrate = self.runge_kutta_step
        self.type = None

    def connect(self, system_output, output_index=0, input_index=0):
        self.inputs[input_index] = system_output
        self.connections[input_index] = output_index

    @staticmethod
    def dummy_input(t=None):
        return [t*0]

    def update_input(self):
        u = []
        for i in range(len(self.inputs)):
            inputs = self.inputs[i](self.t)
            index = self.connections[i]
            u.append(inputs[index])
        self.u = np.array(u)

    def update_output(self):
        return self.y

    def update_state(self):
        return self.x

    def output(self, time=None, index=0):
        return self.y

    def continuous_step(self, dt=0.0):
        self.dt = dt
        self.integrate()
        self.save_step()

    def discrete_step(self):
        self.k += 1
        self.t = self.k * self.dt
        self.update_input()
        self.x = self.update_state()
        self.y = self.update_output()
        self.save_step()

    def euler_step(self):
        self.t = self.t + self.dt
        self.update_input()
        self.x = self.x + self.dt * self.update_state()
        return self.x

    def runge_kutta_step(self):
        self.t = self.t + self.dt
        self.update_input()
        a = self.update_state()
        self.x = self.x + self.dt / 2 * a
        b = self.update_state()
        self.x = self.x + self.dt / 2 * b
        c = self.update_state()
        self.x = self.x + self.dt * c
        d = self.update_state()
        self.x = self.x + self.dt / 6 * (a + 2 * b + 2 * c + d)

    def save_step(self):
        self.archive_t.append(self.t)
        self.archive_u.append(self.u)
        self.archive_x.append(self.x)
        self.archive_y.append(self.update_output())

    def get_state(self):
        table = []
        for i in range(len(self.x)):
            x = []
            for item in self.archive_x:
                x.append(item[i])
            table.append(x)
        return table

    def plot(self, signal='output', index=None):
        # t = self.dt * np.linspace(0, len(self.archive_y), len(self.archive_y))
        # print(self.archive_time)
        # pl.figure()

        pl.grid(True)
        if signal == 'output':
            y = []

            if type(self.y) == np.array:
                n = len(self.y)
            else:
                n = 1
            for i in range(n):
                y.append([])
                for item in self.archive_y:
                    if n == 1:
                        y[i].append(item)
                    else:
                        y[i].append(item[i])

                if index is None:
                    pl.plot(self.archive_t, y[i], label='y')
                else:
                    if i == index:
                        pl.plot(self.archive_t, y[i], label='y')
        pl.ylabel('y[-]')

        if signal == 'input':
            for u in self.archive_u:
                pl.plot(self.archive_t, u)
            pl.ylabel('u[-]')

        if signal == 'state':
            for i in range(len(self.x)):
                x = []
                for item in self.archive_x:
                    x.append(item[i])
                pl.plot(self.archive_t, x, label='x_{0}'.format(i))

            pl.ylabel('x[-]')
            pl.legend()

        pl.xlabel('t[s]')


class Gain(System):

    def __init__(self, model=None, gain=1.0):
        super().__init__()
        self.gain = gain
        self.model = model

    def output(self, time=None, index=0):
        self.update_input()
        self.archive_t.append(time)
        self.archive_y.append(self.u * self.gain)
        return self.u * self.gain


class LTISystem(System):
    def __init__(self, model=None, matrix_a=None, matrix_b=None,
                 matrix_c=None, matrix_d=None, x_0=None, dt=0.0):

        self.matrix_a = np.array(matrix_a)
        self.matrix_b = np.array(matrix_b)
        self.matrix_c = np.array(matrix_c)
        self.matrix_d = np.array(matrix_d)
        self.n_inputs = len(matrix_b[0])
        self.n_outputs = len(matrix_c)
        self.n = len(matrix_a)
        self.u = np.zeros(self.n_inputs)

        super().__init__(model, x_0, n_inputs=self.n_inputs, n=self.n, dt=dt)

    def observability_matrix(self):
        n = len(self.x)
        h_o = np.zeros([n, n])
        for i in range(n):
            h_o[i, :] = self.matrix_c[:]
            for j in range(i):
                h_o[i, :] = h_o[i, :] @ self.matrix_a
        return h_o


class ContinuousLTISystem(LTISystem):
    def __init__(self, model=None, matrix_a=None, matrix_b=None,
                 matrix_c=None, matrix_d=None, x_0=None):

        self.type = 'continuous'
        super().__init__(model, matrix_a, matrix_b, matrix_c, matrix_d, x_0)
        self.y = self.update_output()
        self.save_step()

    def update_state(self):
        self.update_input()
        dx = self.matrix_a @ self.x + self.matrix_b @ self.u
        return dx

    def update_output(self):
        self.y = self.matrix_c @ self.x + self.matrix_d @ self.u
        return self.y


class DiscreteLTISystem(LTISystem):
    def __init__(self, model, matrix_a, matrix_b, matrix_c, matrix_d, x_0=None, dt=0.0):
        super().__init__(model, matrix_a, matrix_b, matrix_c, matrix_d, x_0, dt)
        self.type = 'discrete'
        self.dt = dt
        self.x = np.array(x_0)
        self.save_step()

    def update_state(self):
        self.update_input()
        self.x = self.matrix_a @ self.x + self.matrix_b @ self.u
        return self.x

    def update_output(self):
        self.update_input()
        self.y = self.matrix_c @ self.x + self.matrix_d @ self.u
        return self.y


class Sum(System):
    def __init__(self, model: Model, signs):
        self.n_inputs = len(signs)
        super().__init__(model, n_inputs=self.n_inputs)
        self.type = 'continuous'
        self.signs = signs
        self.archive_y.append([0])
        self.archive_t.append(0)
        self.y = 0

    def output(self, time=None, index=0):
        output = 0
        for i in range(len(self.inputs)):
            output += self.signs[i] * self.inputs[i](time)[0]

        self.archive_y.append([output])
        self.y = [output]
        self.archive_t.append(self.model.t)
        return output

    def continuous_step(self, time):
        self.output(time, 0)


class PI(System):
    def __init__(self, model: Model, p, i, x_0=None):
        super().__init__(model, x_0=x_0)
        self.y = self.x
        self.i = i
        self.p = p
        self.save_step()

    def continuous_step(self, time):
        self.update_input()
        self.x = self.x + self.i * self.u
        self.y = self.x + self.u * self.p
        self.t = time
        self.save_step()


class KalmanFilter(DiscreteLTISystem):
    def __init__(self, model: Model, controlled_system: DiscreteLTISystem, dt, error_obs, error_proc):

        self.n = len(controlled_system.x)
        self.controlled_system = controlled_system
        n_inputs = controlled_system.n_inputs + len(controlled_system.matrix_c)

        x0 = [-600, 3e3, 100, 500]
        super().__init__(model, controlled_system.matrix_a, controlled_system.matrix_b,
                         controlled_system.matrix_c, controlled_system.matrix_d, x_0=x0, dt=dt)

        for i in range(n_inputs):
            self.inputs.append(self.dummy_input)
            self.connections.append(0)

        self.error_proc = error_proc
        self.error_obs = error_obs

        self.P = self.covariance(self.error_proc)
        self.R = self.covariance(self.error_obs)
        self.Q = np.zeros(self.n)


    @staticmethod
    def covariance(sigma):
        cov_matrix = np.eye(len(sigma))
        for i in range(len(sigma)):
            cov_matrix[i, i] = sigma[i] ** 2
        return cov_matrix

    def update_output(self):
        # self.y = self.x
        return self.y

    def discrete_step(self):
        u = self.inputs[0](self.t)

        # Prediction steps
        self.x = self.matrix_a @ self.x + self.matrix_b @ u

        matrix_p = self.matrix_a @ self.P @ self.matrix_a.T + self.Q
        self.P = np.eye(self.n)

        for i in range(self.n):
            self.P[i, i] = matrix_p[i, i]  # Takes only diagonal

        matrix_s = self.R + self.matrix_c @ self.P @ self.matrix_c.T
        matrix_k = self.P @ self.matrix_c.T @ np.linalg.inv(matrix_s)

        # Difference between real output and estimated output
        self.k = self.k + 1
        self.t = self.k * self.dt
        y_meas = np.array(self.inputs[1](self.t))
        y = self.matrix_c @ self.x

        # Update equation
        self.x = self.x + matrix_k @ (y_meas - y)

        matrix_i = np.eye(self.n)
        self.P = (matrix_i - matrix_k @ self.matrix_c) @ self.P
        self.y = self.matrix_c @ self.x
        self.save_step()


def unit_step(t):
    if t >= 0:
        return [1]
    else:
        return [0]


def unit_impulse(t):
    if t == 0:
        return 1
    else:
        return 0


def programme_control(t):
    u = 5.5
    if t > 4:
        u = u * 0.6
    if t > 7:
        u = u * 0.35
    if t > 10:
        u = u * 0.1
    if t > 20:
        u = u * 0.1
    if t > 25:
        u = u * 1.4
    return [u + (np.random.rand()-0.5)]


def zeros():
    return 0


def white_noise(t):
    return [(np.random.rand() - 0.5) * 0]







import os
import numpy as np
import pylab as pl
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
from matplotlib import rc

import logging
mpl_logger = logging.getLogger('matplotlib')
mpl_logger.setLevel(logging.WARNING)


class Results:
    MINIMIZE = -1
    MAXIMIZE = 1

    def __init__(self, problem):
        self.problem = problem

    def value(self, individual, name):
        # TODO: check speed
        val = 0

        try:
            # try costs
            # get the index of the required parameter
            index = 0  # default - one parameter
            for i in range(len(self.problem.costs)):
                cost = self.problem.costs[i]
                if cost['name'] == name:
                    index = i
                    break

            # ind = list(self.problem.costs.keys()).index(name)
            val = individual.costs[index]
        except ValueError:
            # try parameters
            ind = list(self.problem.parameters.keys()).index(name)
            val = individual.vector[ind]
        except ValueError:
            self.problem.logger.error("Value '{}' not found".format(name))

        return val

    def find_minimum(self, name=None):
        """
        Search the optimal value for the given (by name parameter) single objective .

        :param name:
        :return:
        """
        # get the index of the required parameter
        index = 0  # default - one parameter
        if name:
            for i in range(len(self.problem.costs)):
                cost = self.problem.costs[i]
                if cost['name'] == name:
                    index = i
        #if len(self.problem.data_store.individuals) is not 0:
        #    min_l = [min(self.problem.data_store.individuals, key=lambda x: x.costs[index])]
        #else:
        if len(self.problem.data_store.populations[-1].archives) > 0:
            min_l = [min(self.problem.data_store.populations[-1].archives, key=lambda x: x.costs[index])]
        else:
            if len(self.problem.data_store.populations[-1].individuals) is not 0:
                min_l = [min(self.problem.data_store.populations[-1].individuals, key=lambda x: x.costs[index])]
        # for population in self.problem.data_store.populations:
        opt = min(min_l, key=lambda x: x.costs[index])
        return opt

    def find_pareto(self, name1, name2):
        pareto_front_x = []
        pareto_front_y = []
        for population1 in self.problem.data_store.populations:
            for individual1 in population1.individuals:
                is_pareto = True

                for population2 in self.problem.data_store.populations:
                    for individual2 in population2.individuals:
                        # TODO: MINIMIZE and MAXIMIZE
                        if self.value(individual1, name1) > self.value(individual2, name1) \
                                and self.value(individual1, name2) > self.value(individual2, name2):
                            is_pareto = False

                if is_pareto:
                    pareto_front_x.append(self.value(individual1, name1))
                    pareto_front_y.append(self.value(individual1, name2))

        return pareto_front_x, pareto_front_y

    def pareto_values(self):
        """
        :return: a list of lists which contains the optimal values of the cost function:
                l_sol[[c11, c12, ... c1n], ... [cm1, cm2, ... cmn]]
        """

        population = self.problem.data_store.populations[-1]
        l_sol = []

        if len(self.problem.data_store.populations[-1].archives) < 1:
            if len(population.individuals) > 1:
                for individual in population.individuals:
                    l_sol.append(individual.costs)
        else:
            if len(population.archives) > 1:
                for individual in population.archives:
                    l_sol.append(individual.costs)
        return l_sol

    def parameters(self):
        out = []
        for population in self.problem.data_store.populations:
            for individual in population.individuals:
                out.append(individual.vector)
        return out

    def costs(self):
        out = []
        for population in self.problem.data_store.populations:
            for individual in population.individuals:
                out.append(individual.costs)
        return out

class GraphicalResults(Results):

    def __init__(self, problem):
        super().__init__(problem)
        self.name = ""
        # rc('font', **{'family': 'sans-serif', 'sans-serif': ['Helvetica']})
        # pl.rcParams['figure.figsize'] = 6, 4
        rc('text', usetex=True)
        self.labels_size = 16
        self.tick_size = 12

    def plot_scatter(self, name1, name2, filename=None, population_number=None):
        figure = Figure()
        figure.clf()

        if population_number is None:
            populations = self.problem.data_store.populations
        else:
            populations = [self.problem.data_store.populations[population_number]]

        for population in populations:
            values1 = []
            values2 = []
            for individual in population.individuals:
                values1.append(self.value(individual, name1))
                values2.append(self.value(individual, name2))
            pl.scatter(values1, values2)

        # pareto front
        values1 = []
        values2 = []
        population = populations[-1]
        for individual in population.individuals:
            if individual.front_number == 1:
                values1.append(self.value(individual, name1))
                values2.append(self.value(individual, name2))
            pl.scatter(values1, values2, c='k')

        ax = pl.gca()
        # ax.set_yscale('log')
        # ax.set_xscale('log')
        #ax.set_xlim(-1.5, -0.9)
        #ax.set_ylim(0.3, 0.8)
        #ax.set_xlim(0.00000, 0.0004)
        #ax.set_ylim(50, 200)
        #ax.set_xlim(0.00000, 0.001)
        #ax.set_ylim(60, 300)
        ax.set_xlim(0.00000, 0.00125)
        ax.set_ylim(-0.00001, 0.0002)

        # labels
        pl.grid()
        pl.xlabel("${}$".format(name1))
        pl.ylabel("${}$".format(name2))

        if filename is not None:
            pl.savefig(filename)
        else:
            pl.savefig(self.problem.working_dir + os.sep + "scatter.pdf")
        pl.close()

    def plot_convergence_chart(self, name1, filename=None, population_number=None):

        figure = Figure()
        figure.clf()

        # all individuals
        if population_number is None:
            populations = self.problem.data_store.populations
        else:
            populations = [self.problem.data_store.populations[population_number]]

        results = []
        for population in populations:
            min_l = min(population.individuals, key=lambda x: x.costs[0])
            results.append(min_l.costs)

        pl.grid()
        pl.xlabel("Number of generation")
        pl.ylabel("TOC [eur]".format(name1))

        pl.plot(results)
        if filename is not None:
            pl.savefig(filename)
        else:
            pl.savefig("convergence.pdf")
        pl.close()

    def plot_scatter_vectors(self, name1, name2, filename=None, population_number=None):
        figure = Figure()
        figure.clf()

        # all individuals
        if population_number is None:
            populations = self.problem.data_store.populations
        else:
            populations = [self.problem.data_store.populations[population_number]]

        for population in populations:
            values1 = []
            values2 = []
            vector_x = []
            vector_y = []
            for individual in population.individuals:
                values1.append(individual.vector[0])
                values2.append(individual.vector[1])
                vector_x.append(individual.velocity_i[0])
                vector_y.append(individual.velocity_i[1])

            pl.scatter(values1, values2)
            pl.quiver(values1, values2, vector_x, vector_y)

        # pareto front
        values1 = []
        values2 = []
        population = populations[-1]
        for individual in population.individuals:
            if individual.front_number == 1:
                values1.append(self.value(individual, name1))
                values2.append(self.value(individual, name2))
        pl.scatter(values1, values2, c='k')

        # labels
        pl.grid()
        pl.xlabel("${}$".format(name1))
        pl.ylabel("${}$".format(name2))

        if filename is not None:
            pl.savefig(filename)
        else:
            pl.savefig(self.problem.working_dir + os.sep + "scatter.pdf")
        pl.close()

    def plot_individuals(self, name, filename=None):
        # all individuals
        n = 1
        for population in self.problem.data_store.populations:
            ind = []
            values = []
            for individual in population.individuals:
                ind.append(n + len(ind))
                values.append(self.value(individual, name))
            n += len(ind)
            pl.scatter(ind, values)

        # labels
        pl.grid()
        pl.xlabel("$N$")
        pl.ylabel("${}$".format(name))
        # pl.ylim(0, 0.0003)

        if filename is not None:
            pl.savefig(filename)
        else:
            pl.savefig(self.problem.working_dir + os.sep + "individuals.pdf")
        pl.close()

    def plot_all_individuals(self, filename=None):
        for population in self.problem.data_store.populations:
            for individual in population.individuals:
                pl.plot(individual.vector[0], 'x')

        if filename is not None:
            pl.savefig(filename)
        else:
            pl.savefig(self.problem.working_dir + "all_individuals.pdf")

    def plot_populations(self):
        for population in self.problem.data_store.populations:
            figure_name = self.problem.working_dir + "pareto_" + \
                          str(self.problem.data_store.populations.index(population)) + ".pdf"
            if len(population.individuals) > 1:
                figure = Figure()
                FigureCanvas(figure)
                plot = figure.add_subplot(111)
                plot.tick_params(axis='both', which='major', labelsize=self.tick_size)
                plot.tick_params(axis='both', which='minor', labelsize=self.tick_size - 2)
                ax = figure.axes[0]
                colors = ['red', 'green', 'blue', 'yellow', 'purple', 'black']

                for individual in population.individuals:
                    # TODO: Make for more objective values
                    #

                    if hasattr(individual, 'front_number'):
                            scale = 100 / (individual.front_number / 1.)
                            ax.scatter(individual.costs[0], individual.costs[1],
                                       scale, c=colors[(individual.front_number - 1) % 6])
                labels = []
                for cost in self.problem.costs:
                    labels.append(cost['name'])
                x_label = r'$' + labels[0] + '$'
                y_label = r'$' + labels[1] + '$'
                ax.set_xlabel(x_label)
                ax.set_ylabel(y_label)
                ax.grid()

                figure.savefig(figure_name)

    def plot_gradients(self):
        population = self.problem.data_store.populations[-1]  # Last population
        figure_name = self.problem.working_dir + "gradients_" + str(population.number) + ".pdf"
        if len(population.individuals) > 1:
            figure = Figure()
            FigureCanvas(figure)
            plot = figure.add_subplot(111)
            plot.tick_params(axis='both', which='major', labelsize=self.tick_size)
            plot.tick_params(axis='both', which='minor', labelsize=self.tick_size - 2)
            ax = figure.axes[0]
            colors = ['red', 'green', 'blue', 'yellow', 'purple', 'black']

            grad = [[], []]

            sorted(population.individuals, key=lambda item: item.costs[0])

            for individual in population.individuals:
                # TODO: Make for more objective values

                if hasattr(individual, 'front_number'):
                    if individual.front_number == 1:
                        gradient_length = 0

                        for i in range(2):
                            for x in individual.gradient:
                                gradient_length += x[i] ** 2

                            gradient_length = np.sqrt(gradient_length)
                            grad[i].append(gradient_length)

            x_axis = range(1, len(grad[0]) + 1)
            ax.bar(x_axis, grad[0], align='edge', width=0.45)
            ax.bar(x_axis, grad[1], align='center', width=0.45)
            ax.set_xlabel(r'$i$ [-]', size=self.labels_size)
            ax.set_ylabel(r'$\| \nabla F_1 \|, \| \nabla F_2 \|$', size=self.labels_size)
            ax.grid()
            pl.show()
            pl.tight_layout()
            figure.savefig(figure_name)



import math

import unittest
from scipy import integrate

from artap.problem import Problem
from artap.algorithm_nlopt import NLopt, LN_BOBYQA
from artap.algorithm_bayesopt import BayesOptSerial
from artap.algorithm_genetic import NSGAII
from artap.algorithm_sweep import SweepAlgorithm
from artap.benchmark_functions import Booth
from artap.results import Results
from artap.operators import LHSGeneration
from artap.surrogate_scikit import SurrogateModelScikit

from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import ExtraTreesRegressor


class MyProblemCoil(Problem):
    def set(self):
        self.parameters = [{'name': 'x1', 'initial_value': 0.01, 'bounds': [5e-3, 50e-3]},
                      {'name': 'x2', 'initial_value': 0.01, 'bounds': [5e-3, 50e-3]},
                      {'name': 'x3', 'initial_value': 0.01, 'bounds': [5e-3, 50e-3]},
                      {'name': 'x4', 'initial_value': 0.01, 'bounds': [5e-3, 50e-3]},
                      {'name': 'x5', 'initial_value': 0.01, 'bounds': [5e-3, 50e-3]},
                      {'name': 'x6', 'initial_value': 0.01, 'bounds': [5e-3, 50e-3]},
                      {'name': 'x7', 'initial_value': 0.01, 'bounds': [5e-3, 50e-3]},
                      {'name': 'x8', 'initial_value': 0.01, 'bounds': [5e-3, 50e-3]},
                      {'name': 'x9', 'initial_value': 0.01, 'bounds': [5e-3, 50e-3]},
                      {'name': 'x10', 'initial_value': 0.01, 'bounds': [5e-3, 50e-3]}]

    def intl22(self, R2, R, dZ, phi):
        return math.sqrt(R2 ** 2 + R ** 2 - 2.0 * R2 * R * math.cos(phi) + dZ ** 2)

    def intg(self, R2, R, dZ):
        # div J = 0 - nonconstant current density
        f = lambda phi: math.log(R2 - R * math.cos(phi) + self.intl22(R2, R, dZ, phi)) * math.cos(phi)
        return integrate.quad(f, 0, 2.0 * math.pi, epsabs=1e-3, epsrel=1e-3)[0]

    def inth(self, R2, R, dZ):
        # div J = 0 - nonconstant current density
        f = lambda phi: - math.log(dZ + self.intl22(R2, R, dZ, phi))
        return integrate.quad(f, 0, 2.0 * math.pi, epsabs=1e-3, epsrel=1e-3)[0]

    def integral(self, rc, zc, R, Z):
        w = 0.001
        h = 0.0015

        R1 = rc
        R2 = rc + w
        Z1 = zc
        Z2 = zc + h

        mu0 = 4.0 * math.pi * 1e-7
        Jext = 2e6
        # div J = 0 - nonconstant current density
        C = mu0 * Jext * w * h / (4 * math.pi * (Z2 - Z1) * math.log(R2 / R1))

        # upper coil
        Bru = C * (self.intg(R2, R, Z2 - Z) - self.intg(R2, R, Z1 - Z) - self.intg(R1, R, Z2 - Z) + self.intg(R1, R, Z1 - Z))
        Bzu = C * (self.inth(R2, R, Z2 - Z) - self.inth(R2, R, Z1 - Z) - self.inth(R1, R, Z2 - Z) + self.inth(R1, R, Z1 - Z))

        # lower coil
        Brl = C * (self.intg(R2, R, -Z1 - Z) - self.intg(R2, R, -Z2 - Z) - self.intg(R1, R, -Z1 - Z) + self.intg(R1, R, -Z2 - Z))
        Bzl = C * (self.inth(R2, R, -Z1 - Z) - self.inth(R2, R, -Z2 - Z) - self.inth(R1, R, -Z1 - Z) + self.inth(R1, R, -Z2 - Z))

        return [Bru + Brl, Bzu + Bzl]

    def integral_all(self, R, Z, x):
        Br = 0.0
        Bz = 0.0

        for k in range(0, 9):
            rc = x[k]
            zc = k * 1.5e-3

            B = self.integral(rc, zc, R, Z)

            Br = Br + B[0]
            Bz = Bz + B[1]

        return [Br, Bz]

    def evaluate(self, x):
        pass


class MyProblemCoilOne(MyProblemCoil):

    def evaluate(self, individual):
        x = individual.vector
        B0 = 2e-3

        dxy = 0.5e-3
        nx = 8
        ny = 8
        dx = (5e-3 - dxy) / (nx - 1)
        dy = (5e-3 - dxy) / (ny - 1)

        f1 = 0.0
        for i in range(0, nx):
            xx = dxy + i * dx
            for j in range(0, ny):
                yy = dxy + j * dy

                [Br, Bz] = self.integral_all(xx, yy, x)
                Bp1s = math.sqrt((Br - 0.0)**2 + (Bz - B0)**2)
                f1 = max(f1, Bp1s)

        print("value = {}, \tparams = {}".format([f1], x))
        return [f1]


class MyProblemCoilMultiTwo1(MyProblemCoil):

    def evaluate(self, x):
        B0 = 2e-3

        dxy = 0.5e-3
        nx = 8
        ny = 8
        dx = (5e-3 - dxy) / (nx - 1)
        dy = (5e-3 - dxy) / (ny - 1)

        f1 = 0.0
        f2 = sum(x) * 1e3
        for i in range(0, nx):
            xx = dxy + i * dx
            for j in range(0, ny):
                yy = dxy + j * dy

                [Br, Bz] = self.integral_all(xx, yy, x)
                Bp1s = math.sqrt((Br - 0.0)**2 + (Bz - B0)**2)
                f1 = max(f1, Bp1s)

        return [1e3 * f1, 1e3 * f2]


class MyProblemCoilMultiTwo2(MyProblemCoil):
    def __init__(self, name):
        super().__init__(name, costs=['F1', 'F2'])

    def evaluate(self, individual):
        x = individual.vector
        B0 = 2e-3

        dxy = 0.5e-3
        nx = 8
        ny = 8
        dx = (5e-3 - dxy) / (nx - 1)
        dy = (5e-3 - dxy) / (ny - 1)

        f1 = 0.0
        f3 = 0.0
        for i in range(0, nx):
            xx = dxy + i * dx
            for j in range(0, ny):
                yy = dxy + j * dy

                [Br, Bz] = self.integral_all(xx, yy, x)
                Bp1s = math.sqrt((Br - 0.0)**2 + (Bz - B0)**2)
                f1 = max(f1, Bp1s)

                dxsi = 0.5e-3
                [Brp, Bzp] = self.integral_all(xx + dxsi, yy, x)
                [Brm, Bzm] = self.integral_all(xx - dxsi, yy, x)
                Bp3 = math.sqrt((Brp - Br) ** 2 + (Bzp - Bz) ** 2) + math.sqrt((Brm - Br) ** 2 + (Bzm - Bz) ** 2)
                f3 = max(f3, Bp3)

        return [1e3 * f1, 1e3 * f3]


class MyProblemCoilMultiThree(MyProblemCoil):
    def __init__(self, name):
        super().__init__(name, costs=['F1', 'F2', 'F3'])

    def evaluate(self, individual):
        x = individual.vector
        B0 = 2e-3

        dxy = 0.5e-3
        nx = 8
        ny = 8
        dx = (5e-3 - dxy) / (nx - 1)
        dy = (5e-3 - dxy) / (ny - 1)

        f1 = 0.0
        f2 = sum(x)*1e3
        f3 = 0.0
        for i in range(0, nx):
            xx = dxy + i * dx
            for j in range(0, ny):
                yy = dxy + j * dy

                [Br, Bz] = self.integral_all(xx, yy, x)
                Bp1s = math.sqrt((Br - 0.0)**2 + (Bz - B0)**2)
                f1 = max(f1, Bp1s)

                dxsi = 0.5e-3
                [Brp, Bzp] = self.integral_all(xx + dxsi, yy, x)
                [Brm, Bzm] = self.integral_all(xx - dxsi, yy, x)
                Bp3 = math.sqrt((Brp - Br) ** 2 + (Bzp - Bz) ** 2) + math.sqrt((Brm - Br) ** 2 + (Bzm - Bz) ** 2)
                f3 = max(f3, Bp3)

        return [f1, f2, f3]


class MyProblemBooth(Problem):
    """ Describe simple one objective optimization problem. """
    def set(self):
        self.parameters = {'x_1': {'initial_value': 2.5, 'bounds': [-10, 10]},
                           'x_2': {'initial_value': 1.5, 'bounds': [-10, 10]}}

        self.costs = [{'name': 'F'}]

    def evaluate(self, x):
        return [Booth.eval(x)]


class TestSimpleOptimization(unittest.TestCase):
    """ Tests optimization problem."""

    def xtest_local_problem_booth(self):
        problem = MyProblemBooth("MyProblemBooth")

        problem.surrogate = SurrogateModelScikit(problem)
        #kernel = 1.0 * RationalQuadratic(length_scale=1.0, alpha=0.1)
        #problem.surrogate.regressor = GaussianProcessRegressor(kernel=kernel)
        #problem.surrogate.has_epsilon = True

        problem.surrogate.regressor = ExtraTreesRegressor(n_estimators=10)
        # problem.surrogate.regressor = DecisionTreeRegressor()

        problem.surrogate.train_step = 50
        problem.surrogate.score_threshold = 0.0

        # sweep analysis (for training)
        gen = LHSGeneration(problem.parameters)
        gen.init(problem.surrogate.train_step)
        algorithm_sweep = SweepAlgorithm(problem, generator=gen)
        algorithm_sweep.run()

        # optimization
        algorithm = NLopt(problem)
        algorithm.options['algorithm'] = LN_BOBYQA
        algorithm.options['n_iterations'] = 200
        algorithm.run()

        problem.logger.info("surrogate.predict_counter: {}".format(problem.surrogate.predict_counter))
        problem.logger.info("surrogate.eval_counter: {}".format(problem.surrogate.eval_counter))

        # print(problem.surrogate.x_data)
        # print(problem.surrogate.y_data)

        results = Results(problem)
        optimum = results.find_minimum('F')
        self.assertAlmostEqual(optimum, 1e-6, 3)
        """
        kernels = [1.0 * RBF(length_scale=1.0, length_scale_bounds=(1e-1, 10.0)),
                   1.0 * RationalQuadratic(length_scale=1.0, alpha=0.1),
                   1.0 * ExpSineSquared(length_scale=1.0, periodicity=3.0,
                                        length_scale_bounds=(0.1, 10.0),
                                        periodicity_bounds=(1.0, 10.0)),
                   ConstantKernel(0.1, (0.01, 10.0))
                   * (DotProduct(sigma_0=1.0, sigma_0_bounds=(0.1, 10.0)) ** 2),
                   1.0 * Matern(length_scale=1.0, length_scale_bounds=(1e-5, 1e5), nu=1.5)]
        
        for kernel in kernels:
            print(kernel)

            problem.surrogate = SurrogateModelScikit(problem)
            # problem.surrogate.regressor = GaussianProcessRegressor(kernel=kernel)
            # set threshold
            problem.surrogate.sigma_threshold = 0.1
            problem.surrogate.train_step = 10

            algorithm = NLopt(problem)
            algorithm.options['algorithm'] = LN_BOBYQA
            algorithm.options['n_iterations'] = 200
            algorithm.run()

            problem.logger.info("surrogate.predict_counter: {}".format(problem.surrogate.predict_counter))
            problem.logger.info("surrogate.eval_counter: {}".format(problem.surrogate.eval_counter))
        """

    def xtest_local_problem_coil_one(self):
        problem = MyProblemCoilOne("MyProblemCoilOne")

        # enable surrogate
        problem.surrogate = SurrogateModelScikit(problem)
        problem.surrogate.regressor = DecisionTreeRegressor()
        problem.surrogate.train_step = 30
        problem.surrogate.score_threshold = 0.0

        # sweep analysis (for training)
        gen = LHSGeneration(problem.parameters)
        gen.init(problem.surrogate.train_step)
        algorithm_sweep = SweepAlgorithm(problem, generator=gen)
        algorithm_sweep.run()

        # optimization
        algorithm = NLopt(problem)
        algorithm.options['algorithm'] = LN_BOBYQA
        algorithm.options['n_iterations'] = 50
        algorithm.run()

        problem.logger.info("surrogate.predict_counter: {}".format(problem.surrogate.predict_counter))
        problem.logger.info("surrogate.eval_counter: {}".format(problem.surrogate.eval_counter))

        results = Results(problem)
        optimum = results.find_minimum('F1')
        self.assertAlmostEqual(optimum, 5e-5, 4)

    def xtest_local_problem_coil_one_bobyqa_optimum(self):
        problem = MyProblemCoilOne("MyProblemCoilOne")

        # optimization
        algorithm = NLopt(problem)
        algorithm.options['algorithm'] = LN_BOBYQA
        algorithm.options['n_iterations'] = 500
        algorithm.run()

        problem.logger.info("surrogate.predict_counter: {}".format(problem.surrogate.predict_counter))
        problem.logger.info("surrogate.eval_counter: {}".format(problem.surrogate.eval_counter))

        results = Results(problem)
        optimum = results.find_minimum('F1')
        print("BOBYQA = {}".format(optimum))
        # Bayes = 3.846087978861188e-05
        # self.assertAlmostEqual(optimum, 1e-6, 4)

    def xtest_local_problem_coil_one_bayesopt_optimum(self):
        problem = MyProblemCoilOne("MyProblemCoilOne")

        # optimization
        algorithm = BayesOptSerial(problem)
        algorithm.options['n_iterations'] = 500
        algorithm.run()

        problem.logger.info("surrogate.predict_counter: {}".format(problem.surrogate.predict_counter))
        problem.logger.info("surrogate.eval_counter: {}".format(problem.surrogate.eval_counter))

        results = Results(problem)
        optimum = results.find_minimum('F1')
        print("Bayes = {}".format(optimum))
        # Bayes = 4.347142168223674e-05
        # self.assertAlmostEqual(optimum, 1e-6, 4)

    def xtest_local_problem_coil_one_nsgaii_optimum(self):
        problem = MyProblemCoilOne("MyProblemCoilOne")

        # optimization
        algorithm = NSGAII(problem)
        algorithm.options['max_population_number'] = 100
        algorithm.options['max_population_size'] = 50
        algorithm.run()

        problem.logger.info("surrogate.predict_counter: {}".format(problem.surrogate.predict_counter))
        problem.logger.info("surrogate.eval_counter: {}".format(problem.surrogate.eval_counter))

        results = Results(problem)
        optimum = results.find_minimum('F1')
        print("NSGAII = {}".format(optimum))
        # NSGAII = 8.099681801799041e-06
        # self.assertAlmostEqual(optimum, 1e-6, 4)

"""
class TestSimpleOptimization(unittest.TestCase):
    def test_local_problem_booth(self):
        problem = MyProblemCoilOne("LocalPythonProblem")
        #problem = MyProblemMultiTwo2("LocalPythonProblem")

        #algorithm = BayesOptSerial(problem)
        #algorithm.options['verbose_level'] = 0
        #algorithm.options['n_iterations'] = 100

        algorithm = NLopt(problem)
        algorithm.options['algorithm'] = LN_BOBYQA
        algorithm.options['n_iterations'] = 200

        #algorithm = NSGA_II(problem)
        #algorithm.options['max_population_number'] = 80
        #algorithm.options['max_population_size'] = 20

        t_s = time.time()

        algorithm.run()

        t = time.time() - t_s
        print('Elapsed time:', t)

        print("surrogate_predict_counter: ", problem.surrogate_predict_counter)
        print("surrogate_eval_counter: ", problem.surrogate_eval_counter)

        results = Results(problem)
        optimum = results.find_minimum('F1')
        print(optimum)
        self.assertAlmostEqual(optimum, 1e-6, 3)

"""
"""
def figures(name):
    import matplotlib
    matplotlib.use('Agg')

    import pylab as pl
    from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
    from matplotlib.figure import Figure
    from matplotlib import rc

    data_store = SqliteDataStore(database_file=name + ".sqlite")
    problem = ProblemSqliteDataStore(data_store)

    data_x = []
    data_y = []
    pareto_front_x = []
    pareto_front_y = []
    for population in problem.populations:
        if len(population.individuals) > 1:
            for individual in population.individuals:
                data_x.append(individual.costs[0])
                data_y.append(individual.costs[1])

    results = GraphicalResults(problem)
    pareto_front_x, pareto_front_y = results.find_pareto({'F1': Results.MINIMIZE, 'F2': Results.MINIMIZE})

    pl.rcParams['figure.figsize'] = 10, 4
    pl.rcParams['legend.fontsize'] = 17
    pl.rcParams['text.usetex'] = True
    pl.rcParams['font.size'] = 20
    pl.rcParams['font.serif'] = "Times"
    pl.figure()
    pl.plot(data_x, data_y, 'o', color='#d0d0d0', markersize=3)
    pl.plot(pareto_front_x, pareto_front_y, 'o', markersize=4, label="Pareto Front")
    pl.xlim(1e-4, 8e-4)
    pl.ylim(0, 1e-3)
    pl.grid(True)
    pl.tight_layout()
    pl.legend(loc="upper right")
    pl.xlabel("$F_1$")
    pl.ylabel("$F_2$")
    pl.savefig(name + ".pdf", dpi=200)
"""

if __name__ == '__main__':
    unittest.main()



from abc import *
from random import uniform
from enum import Enum


class Individual(metaclass=ABCMeta):
    """ Collects information about one point in design space. """

    class State(Enum):
        EMPTY = 0
        EVALUATED = 1
        FAILED = 2

    def __init__(self, vector: list):
        self.vector = vector.copy()
        self.costs = []
        self.state = self.State.EMPTY

    def __repr__(self):
        """ :return: [vector[p1, p2, ... pn]; costs[c1, c2, ... cn]] """

        string = "vector: ["
        for i, number in enumerate(self.vector):
            string += str(number)
            if i < len(self.vector) - 1:
                string += ", "
        string = string[:len(string) - 1]
        string += "]"

        string += "; costs:["
        for i, number in enumerate(self.costs):
            string += str(number)
            if i < len(self.costs) - 1:
                string += ", "
        string += "]"

        return string

    def __eq__(self, other):
        return self.vector == other.vector and self.costs == self.costs

    def __hash__(self):
        return id(self)

    def sync(self, individual):
        self.vector = individual.vector
        self.costs = individual.costs
        self.state = individual.state

class GeneticIndividual(Individual):
    def __init__(self, vector: list):
        super().__init__(vector)

        self.gradient = []
        self.feasible = 0.0  # the distance from the feasibility region in min norm
        self.dominate = set()
        self.domination_counter = 0
        self.front_number = None  # TODO: make better
        self.crowding_distance = 0  # TODO: deprecated? --
        self.depends_on = None  # For calculating gradients
        self.modified_param = -1
        # For particle swarm optimization
        self.velocity_i = [0] * len(vector)  # particle velocity
        self.best_parameters = []  # best position individual
        self.best_costs = []  # best error individual

        for i in range(0, len(self.vector)):
            self.velocity_i.append(uniform(-1, 1))

    def __repr__(self):
        """ :return: [vector[p1, p2, ... pn]; costs[c1, c2, ... cn]] """

        string = "{}, ".format(super().__repr__())

        string += ", front number: {}".format(self.front_number)
        string += ", crowding distance: {}".format(self.crowding_distance)
        string += ", gradient: ["
        for i, number in enumerate(self.gradient):
            string += str(number)
            if i < len(self.vector) - 1:
                string += ", "
        string = string[:len(string) - 1]
        string += "]"

        return string

    def sync(self, individual):
        super().sync(individual)

        self.gradient = individual.gradient
        self.feasible = individual.feasible
        self.dominate = individual.dominate
        self.domination_counter = individual.domination_counter
        self.front_number = individual.front_number
        self.crowding_distance = individual.crowding_distance
        self.depends_on = individual.depends_on
        self.modified_param = individual.modified_param
        # For particle swarm optimization
        self.velocity_i = individual.velocity_i
        self.best_parameters = individual.best_parameters
        self.best_costs = individual.best_costs



from .problem import Problem
from .algorithm import Algorithm
from .population import Population
from .environment import Enviroment
from .job import JobSimple

import time
import numpy as np

import os

import sys
sys.path.append(Enviroment.artap_root + os.sep + "lib" + os.sep)
import bayesopt

from multiprocessing import Process, Pipe, Queue, Manager
# from multiprocessing.managers import BaseManager

_l_type = ['L_FIXED', 'L_EMPIRICAL', 'L_DISCRETE', 'L_MCMC', 'L_ERROR']
_sc_type = ['SC_MTL', 'SC_ML', 'SC_MAP', 'SC_LOOCV', 'SC_ERROR']
_surr_name = ["sGaussianProcess", "sGaussianProcessML", "sGaussianProcessNormal", "sStudentTProcessJef", "sStudentTProcessNIG"]


# Python module to get run BayesOpt library in a OO pattern.
# The objective module should inherit this one and override evaluateSample.
class BayesOptContinuous(object):

    # Let's define the vector.
    #
    # For different options: see vector.h and vector.cpp .
    # If a parameter is not defined, it will be automatically set
    # to a default value.
    def __init__(self, n_dim):
        ## Library vector
        self.params = {}
        ## n dimensions
        self.n_dim = n_dim
        ## Lower bounds
        self.lb = np.zeros((self.n_dim,))
        ## Upper bounds
        self.ub = np.ones((self.n_dim,))

    @property
    def parameters(self):
        return self.params

    @parameters.setter
    def parameters(self, params):
        self.params = params

    @property
    def lower_bound(self):
        return self.lb

    @lower_bound.setter
    def lower_bound(self, lb):
        self.lb = lb

    @property
    def upper_bound(self):
        return self.ub

    @upper_bound.setter
    def upper_bound(self, ub):
        self.ub = ub

    ## Function for testing.
    # It should be overriden.
    def evaluateSample(self, x_in):
        raise NotImplementedError("Please Implement this method")

    ## Main function. Starts the optimization process.
    def optimize(self):
        min_val, x_out, error = bayesopt.optimize(self.evaluateSample, self.n_dim,
                                                  self.lb, self.ub,
                                                  self.params)

        return min_val, x_out, error


class BayesOpt(Algorithm):
    """ BayesOpt algorithms """

    def __init__(self, problem: Problem, name="BayesOpt"):
        super().__init__(problem, name)
        self.problem = problem

        self.options.declare(name='l_type', default='L_EMPIRICAL', values=_l_type,
                             desc='Type of learning for the kernel params')
        self.options.declare(name='sc_type', default='SC_MAP', values=_sc_type,
                             desc='Type of learning for the kernel params')
        self.options.declare(name='n_iterations', default=50, lower=1,
                             desc='Maximum BayesOpt evaluations')
        self.options.declare(name='init_method', default=1,
                             desc='Init method') # 1-LHS, 2-Sobol
        self.options.declare(name='n_init_samples', default=10, lower=1,
                             desc='Number of samples before optimization')
        self.options.declare(name='n_iter_relearn', default=10, lower=1,
                             desc='Number of samples before relearn kernel')
        self.options.declare(name='surr_name', default='sGaussianProcessML', values=_surr_name,
                             desc='Name of the surrogate function')
        self.options.declare(name='surr_noise', default=1e-10, lower=0.0,
                             desc='Variance of observation noise')


class BayesOptClassSerial(BayesOptContinuous):
    def __init__(self, algorithm):
        n = len(algorithm.problem.parameters)
        super().__init__(n)

        # algorithm
        self.algorithm = algorithm

        # Size design variables.
        self.lb = np.empty((n,))
        self.ub = np.empty((n,))
        self.params = {}

        self.job = JobSimple(self.algorithm.problem)

    def evaluateSample(self, x):
        return self.job.evaluate_scalar(x)


class BayesOptSerial(BayesOpt):
    """ BayesOpt algorithms """

    def __init__(self, problem: Problem, name="BayesOpt"):
        super().__init__(problem, name)
        self.problem.options['save_level'] = "individual"
        self.bo = BayesOptClassSerial(self)

    def run(self):
        # Figure out bounds vectors.
        i = 0
        for parameter in self.problem.parameters:
            bounds = parameter['bounds']

            self.bo.lb[i] = bounds[0]
            self.bo.ub[i] = bounds[1]

            i += 1

        # set bayesopt
        self.bo.params['n_iterations'] = self.options['n_iterations']
        self.bo.params['n_init_samples'] = self.options['n_init_samples']
        self.bo.params['n_iter_relearn'] = self.options['n_iter_relearn']
        self.bo.params['surr_name'] = self.options['surr_name']
        self.bo.params['surr_noise'] = self.options['surr_noise']
        self.bo.params['init_method'] = self.options['init_method']
        self.bo.params['l_type'] = self.options['l_type']
        self.bo.params['sc_type'] = self.options['sc_type']
        self.bo.params['verbose_level'] = self.options['verbose_level']

        t_s = time.time()
        self.problem.logger.info("BayesOpt: surr_name{}".format(self.options['surr_name']))
        mvalue, x_out, error = self.bo.optimize()
        t = time.time() - t_s
        self.problem.logger.info("BayesOpt: elapsed time: {} s".format(t))

        if error != 0:
            print('Optimization FAILED.')
            print("Error", error)
            print('-' * 35)
        else:
            pass
            # print('Optimization Complete, %f seconds' % (clock() - start))
            # print("Result", x_out, mvalue)
            # print('-' * 35)


class BayesOptClassParallel(Process, BayesOptContinuous):
    def __init__(self, pipe, n, problem: Problem):
        Process.__init__(self)
        BayesOptContinuous.__init__(self, n)

        # output
        self.mvalue = -1.0
        self.x_out = -1.0
        self.error = 0

        self.pipe = pipe
        self.problem = problem

        # Size design variables.
        self.lb = np.empty((n,))
        self.ub = np.empty((n,))
        self.params = {}

    def run(self):
        mvalue, x_out, error = self.optimize()
        self.pipe.send('STOP')

        # set output values
        self.mvalue = mvalue
        self.x_out = x_out
        self.error = error

        # output
        print(self.mvalue)
        print(self.x_out)
        print(self.error)

    def evaluateSample(self, x):
        self.pipe.send(x)
        result = self.pipe.recv()
        return result


def worker(pipe, problem):
    x = None
    while True:
        x = pipe.recv()
        if str(x) == 'STOP':
            break

        job = JobSimple(problem)
        result = job.evaluate_scalar(x)
        pipe.send(result)


class BayesOptParallel(BayesOpt):
    """ BayesOpt algorithms """

    def __init__(self, problem: Problem, name="BayesOpt"):
        super().__init__(problem, name)

        self.pipe_par, self.pipe_child = Pipe()
        self.bo = BayesOptClassParallel(self.pipe_child, len(self.problem.parameters), problem)

    def run(self):
        # Figure out bounds vectors.
        i = 0
        for id in self.problem.parameters:
            bounds = self.problem.parameters[id]['bounds']

            self.bo.lb[i] = bounds[0]
            self.bo.ub[i] = bounds[1]

            i += 1

        # set bayesopt
        self.bo.params['n_iterations'] = self.options['n_iterations']
        self.bo.params['n_init_samples'] = self.options['n_init_samples']
        self.bo.params['n_iter_relearn'] = self.options['n_iter_relearn']
        self.bo.params['surr_name'] = self.options['surr_name']
        self.bo.params['surr_noise'] = self.options['surr_noise']
        self.bo.params['init_method'] = self.options['init_method']
        self.bo.params['l_type'] = self.options['l_type']
        self.bo.params['sc_type'] = self.options['sc_type']
        self.bo.params['verbose_level'] = self.options['verbose_level']

        process = Process(target=worker, args=(self.pipe_par, self.problem))

        self.bo.start()
        process.start()

        self.bo.join()
        process.join()

        print(self.bo.mvalue)
        print(self.bo.x_out)
        print(self.bo.error)

        # self.result = self.mvalue
        """
        if self.bo.error != 0:
            print('Optimization FAILED.')
            print("Error", self.bo.error)
            print('-' * 35)
        else:
            print('Optimization Complete, %f seconds' % (clock() - start))
            print("Result", self.bo.x_out, self.bo.mvalue)
            print('-' * 35)
            
        """